#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	T Marks
# Date:	2020/04/14

from __future__ import absolute_import
from __future__ import print_function
__author__ = "T Marks"
__email__  = "tsmarks@gmail.com"

__name__ = _("Spiral")
__version__= "0.0.1"

from CNC import CNC,Block
from ToolsPage import Plugin
import Utils
from Utils import to_zip
import math
try:
	import Tkinter
	from Queue import *
	from Tkinter import *
	import tkMessageBox
except ImportError:
	import tkinter
	from queue import *
	from tkinter import *
	import tkinter.messagebox as tkMessageBox


import sys   # Trouble Shooting Only!

#==============================================================================
#Spiral class
#==============================================================================
class Spiral:
	def __init__(self,name="Spiral"):
		self.name = name

	#----------------------------------------------------------------------
	def make(self,app, XStart=0.0, YStart=0.0, ZStart=30., AlignAxis="Y", \
			RotAxis="A", StockLeng=20, ReduceDepth=-1, PassDepth=1, \
			Stepover=1, ZApproach=35, SpiralType="Spiral", CutBoth="True", LiftPass="False"):

		#GCode Blocks
		blocks = []
		
		# Load tool and material settings
		toolDiam = CNC.vars['diameter']
		toolRadius = toolDiam / 2.

		#Calc tool diameter with Maximum Step Over allowed
		StepOverInUnitMax = toolDiam * CNC.vars['stepover'] / 100.0		

		#Check parameters
		if RotAxis == "":
			app.setStatus(_("Spiral abort: Rotary Axis is undefined"))
			return

		if SpiralType == "":
			app.setStatus(_("Spiral abort: Spiral Type is undefined"))
			return

		if ZApproach <= ZStart :
			app.setStatus(_("Spiral abort: Approach height must be greater than Z Start"))
			return

		if ReduceDepth > 0 :
			app.setStatus(_("Spiral abort: Depth Reduction must be negative"))
			return

		if Stepover > StepOverInUnitMax and SpiralType == "Spiral":  #if Type is Lines then stepover is degrees not mm
			app.setStatus(_("Spiral abort: Step Over exceeds tool limits"))
			return 
		elif Stepover > StepOverInUnitMax and SpiralType == "Lines":  # This could cause a tool crash, but could also be used to make faceted shapes.
			dr=tkMessageBox.askyesno("Crash Risk","WARNING: Using a larger stepover value than tool's maximum with lines operation may result in a tool crash. Do you want to continue?")
			sys.stdout.write("%s"%(dr))
			if dr == True or dr == "yes" :
				app.setStatus(_("Risk Accepted")) #Using positive logic, if python returns ANYTHING other than True/yes this will not make g-code.  Incase Python uses No instead of False
			else:
				return
		if StockLeng <= 0 :
			app.setStatus(_("Spiral abort: Stock Length to cut must be positive"))
			return
			

		#Add Region disabled to show worked area
		block = Block(self.name + " Outline")
		block.enable = False
		block.append(CNC.grapid(CNC.vars["wx"],CNC.vars["wy"],ZApproach))  ## Cannot trust Safe-Z with 4th axis!!
		if AlignAxis == "X":
			outlineWidth = StockLeng
		else:
			outlineWidth = 0
		if AlignAxis == "Y":
			outlineHeight = StockLeng
		else:
			outlineHeight = 0
		xR,yR = self.RectPath(XStart,YStart,outlineWidth,outlineHeight)
		for x,y in zip(xR,yR):
			block.append(CNC.gline(x,y))
		blocks.append(block)

		
		if StockLeng < toolDiam :
			app.setStatus(_("Spiral abort: Stock Length is too small for this End Mill."))
			return

		#Prepare points for pocketing
		xP=[]
		yP=[]
		rP=[]
		zP=[]
		gP=[]
        

        #---------------------------------------------------------------------
        #Line approach
		if SpiralType == "Lines":
			#Calc number of indexes
			IndexNum = math.ceil(360/Stepover) # Using the step over as Degrees 
						

			#Calc number of pass
			VerticalCount = math.ceil(abs(ReduceDepth) / PassDepth)
			#Calc even depths of cut
			EvenCutDepths = ReduceDepth / VerticalCount
			
			currentR = 0
			currentZ = ZStart - EvenCutDepths
			direction = 1
			if AlignAxis == "X" :
				currentX = XStart + toolRadius
				currentY = YStart
			elif AlignAxis == "Y":
				currentX = XStart
				currentY = YStart + toolRadius
			else:
				app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
				return			
			
			while (currentZ >= (ZStart + ReduceDepth)):
				#sys.stdout.write("~~~~~%s,%s,%s,%s,%s!"%(currentZ,ZStart,ReduceDepth,EvenCutDepths,VerticalCount))
				while (currentR < 360):
					#sys.stdout.write("~~~~~%s,%s,%s,%s,%s!"%(currentR,Stepover,currentX,currentY,VerticalCount))

					#Plunge in
					gP.append(1)
					rP.append(currentR)
					zP.append(currentZ)			
					xP.append(currentX)
					yP.append(currentY)
					if direction == 1:
						if AlignAxis == "X" :
							currentX = StockLeng - toolRadius
							currentY = YStart
						elif AlignAxis == "Y":
							currentX = XStart
							currentY = StockLeng - toolRadius
						else:
							app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
							return
						if CutBoth == "True" :
							direction = -1
					else :
						if AlignAxis == "X" :
							currentX = XStart + toolRadius
							currentY = YStart
						elif AlignAxis == "Y":
							currentX = XStart
							currentY = YStart + toolRadius
						else:
							app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
							return
						direction = 1
					gP.append(1)
					zP.append(currentZ)			
					xP.append(currentX)
					yP.append(currentY)
					rP.append(currentR)
					# Lift before rotating if required, useful to make non-round shape

					if CutBoth == "False" : # Return to start
					
						#Lift Before return
						gP.append(0)
						rP.append(currentR)
						zP.append(ZApproach)
						xP.append(currentX)
						yP.append(currentY)

						#Return to start
						if AlignAxis == "X" :
							currentX = XStart + toolRadius
							currentY = YStart
						elif AlignAxis == "Y":
							currentX = XStart
							currentY = YStart + toolRadius
						else:
							app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
							return
						gP.append(0)
						xP.append(currentX)
						yP.append(currentY)
						rP.append(currentR)
						zP.append(ZApproach)
						#Rotate
						currentR += Stepover
						gP.append(0)
						xP.append(currentX)
						yP.append(currentY)
						rP.append(currentR)
						zP.append(ZApproach)
					elif LiftPass == "True" and CutBoth == "True" :
						gP.append(0)
						rP.append(currentR)
						zP.append(ZApproach)
						xP.append(currentX)
						yP.append(currentY)
						currentR += Stepover
						gP.append(0)
						xP.append(currentX)
						yP.append(currentY)
						rP.append(currentR)
						zP.append(ZApproach)
					elif LiftPass == "False" and CutBoth == "True" :
						currentR += Stepover
				gP.append(0)
				xP.append(currentX)
				yP.append(currentY)
				rP.append(currentR)
				zP.append(ZApproach)
				currentR=0
				gP.append(0)
				xP.append(currentX)
				yP.append(currentY)
				rP.append(currentR)
				zP.append(ZApproach)		

					
					
					
				#Step Down
				currentZ += EvenCutDepths			

		#---------------------------------------------------------------------
        #Spiral approach
		if SpiralType == "Spiral":
			#Calc number of pass
			StepsPerRot = math.ceil(StockLeng/Stepover)
			TotalRot = 360 * StepsPerRot
			
			#Calc steps in depth
			VerticalCount = math.ceil(abs(ReduceDepth) / PassDepth)
			#Calc even depths of cut
			EvenCutDepths = ReduceDepth / VerticalCount			
			
			direction = 1
			currentZ = ZStart - EvenCutDepths
			if AlignAxis == "X" :
				currentX = XStart + toolRadius
				currentY = YStart
			elif AlignAxis == "Y":
				currentX = XStart
				currentY = YStart + toolRadius
			else:
				app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
				return
			currentR = 0
			while (currentZ >= (ZStart + ReduceDepth)):
				
				# Plunge to depth
				currentR += 90 # Ramp the Plunge
				gP.append(1)
				rP.append(currentR)
				zP.append(currentZ)			
				xP.append(currentX)
				yP.append(currentY)
				
				# One Full Rotation for a clean shoulder
				currentR += 360
				gP.append(1)
				rP.append(currentR)
				zP.append(currentZ)			
				xP.append(currentX)
				yP.append(currentY)
				
				
				if AlignAxis == "X" :
					if direction == 1:
						currentX = StockLeng - toolRadius
					else:
						currentX = XStart + toolRadius
					currentY = YStart
				elif AlignAxis == "Y":
					currentX = XStart
					if direction == 1:
						currentY = StockLeng - toolRadius
					else:
						currentY = YStart + toolRadius
				else:
					app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
					return
					
				currentR += TotalRot
				gP.append(1)
				rP.append(currentR)
				zP.append(currentZ)			
				xP.append(currentX)
				yP.append(currentY)
				
				# One Full Rotation for a clean shoulder
				currentR += 360
				gP.append(1)
				rP.append(currentR)
				zP.append(currentZ)			
				xP.append(currentX)
				yP.append(currentY)
				
				if CutBoth == "True" : 
					direction *=  - 1
				else:
					#Retract
					gP.append(0)
					rP.append(currentR)
					zP.append(ZApproach)			
					xP.append(currentX)
					yP.append(currentY)
					# Return and Rewind
					gP.append(0)
					rP.append(currentR)
					zP.append(ZApproach)
					if AlignAxis == "X" :
						currentX = XStart + toolRadius
						currentY = YStart
					elif AlignAxis == "Y":
						currentX = XStart
						currentY = YStart + toolRadius
					else:
						app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
						return
					xP.append(currentX)
					yP.append(currentY)
				
				
				currentZ += EvenCutDepths
				

#Start G-Code Processes
		#Blocks for pocketing
		block = Block(self.name)
		block.append("(Reduce Rotary by Y=%g)"%(ReduceDepth))
		block.append("(Approach: %s )" % (SpiralType))

		#Move safe to first point
		block.append(CNC.grapid(CNC.vars["mx"],CNC.vars["my"],ZApproach))  ## Cannot trust Safe-Z with 4th axis!!
		if AlignAxis == "X" :
			block.append(CNC.grapid(XStart + toolRadius,YStart))
		elif AlignAxis == "Y":
			block.append(CNC.grapid(XStart,YStart + toolRadius))
		else:
			app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
			return
		
		block.append(CNC.zenter(ZApproach))
		block.append(CNC.gcode(1, [("f",CNC.vars["cutfeed"])]))
		
		for g,x,y,z,r in zip(gP, xP,yP,zP, rP):
			if RotAxis == "A" :
				if g==0:
					block.append(CNC.grapidABC(x,y,z,r,CNC.vars["wb"],CNC.vars["wc"]))
					#sys.stdout.write("%s,%s,%s,%s,%s"%(g,x,y,z,r))
				else:
					block.append(CNC.glineABC(x,y,z,r,CNC.vars["wb"],CNC.vars["wc"]))
					#sys.stdout.write("%s,%s,%s,%s,%s"%(g,x,y,z,r))
			elif RotAxis == "B" :
				if g==0:
					block.append(CNC.grapidABC(x,y,z,CNC.vars["wa"],r,CNC.vars["wc"]))
				else:
					block.append(CNC.glineABC(x,y,z,CNC.vars["wa"],r,CNC.vars["wc"]))
			elif RotAxis == "C" :
				if g==0:
					block.append(CNC.grapidABC(x,y,z,CNC.vars["wa"],CNC.vars["wb"],r))
				else:
					block.append(CNC.glineABC(x,y,z,CNC.vars["wa"],CNC.vars["wb"],r))

			
		block.append(CNC.grapid(CNC.vars["wx"],CNC.vars["wy"],ZApproach))  ## Cannot trust Safe-Z with 4th axis!!
		if AlignAxis == "X" :
			block.append(CNC.grapid(XStart + toolRadius,YStart))
		elif AlignAxis == "Y":
			block.append(CNC.grapid(XStart,YStart + toolRadius))
		else:
			app.setStatus(_("Spiral abort: Rotary Axis Not Assigned."))
			return
		block.append(CNC.zexit(ZApproach))
		blocks.append(block)
		tkMessageBox.showinfo("Crash Risk","WARNING: Check CAM file Header for Z move. If it exists, remove it to prevent tool crash.")

		return blocks

	#----------------------------------------------------------------------
	
	def RectPath(self,x,y,w,h):
		xR = []
		yR = []
		xR.append(x)
		yR.append(y)
		xR.append(x + w)
		yR.append(y)
		xR.append(x + w)
		yR.append(y + h)
		xR.append(x)
		yR.append(y + h)
		xR.append(x)
		yR.append(y)
		return (xR,yR)


#==============================================================================
# Spiral Cut on 4th Axis to reduce size
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Reduce Diameter of 4th Axis Stock")

	def __init__(self, master):
		Plugin.__init__(self, master, "Spiral")
		self.icon  = "helical"
		self.group = "CAM"
		self.variables = [
			("name",           "db",    "",  _("Name")),
			("XStart"  ,       "mm",   0.0,  _("X start")),
			("YStart"  ,       "mm",   0.0,  _("Y start")),
			("ZStart" ,        "mm",   30.0, _("Z start")),
			("AlignAxis"  ,      "X,Y",  "Y",  _("Rotary Alignment Axis")),
			("RotAxis"  ,      "A,B,C",  "A",  _("Rotary Axis")),
			("StockLeng"  ,    "mm",   20.0, _("Length of Stock to Reduce")),
			("ReduceDepth"  ,  "mm",    -1.0, _("Depth to Reduce")),
			("PassDepth"  ,    "mm",    1.0, _("Max Depth per Pass")),
			("Stepover"  ,     "mm",    1.0, _("Stepover (spiral=mm, lines=deg)")),
			("ZApproach"  ,    "mm",    35.0, _("Approach Height (Safe Z)")),
			("SpiralType"  , "Spiral,Lines" ,"Spiral", _("Cut Pattern")),
			("CutBoth"     , "True,False","True",_("Cut in Both Directions")),
			("LiftPass"   ,  "True,False","False",_("Lift before rotate"))
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		n = self["name"]
		if not n or n=="default": n="Spiral"
		spiral = Spiral(n)

		blocks = spiral.make(app,
				self.fromMm("XStart"),
				self.fromMm("YStart"),
				self.fromMm("ZStart"),
				self["AlignAxis"],
				self["RotAxis"],
				self.fromMm("StockLeng"),
				self.fromMm("ReduceDepth"),
				self.fromMm("PassDepth"),
				self.fromMm("Stepover"),
				self.fromMm("ZApproach"),
				self["SpiralType"],
				self["CutBoth"],
				self["LiftPass"]
				)

		if blocks is not None:
			active = app.activeBlock()
			if active==0: active=1
			app.gcode.insBlocks(active, blocks, "Spiral")
			app.refresh()
			app.setStatus(_("Spiral: Reduced 4th Axis Stock"))



# SVGcode 0.2
# Converts SVG paths to g-code
# (c) 2018 - Tomas 'Harvie' Mudrunka
# https://github.com/harvie
# License: GPLv2+

# Usage:
# svgcode = SVGcode('./image.svg')
# for path in svgcode.get_gcode():
#	print(path['id'])
#	print(path['path'])

import numpy
from svg_elements import SVG, Path, Move, Shape, Line, Close, Arc


class SVGcode:
	def __init__(self, filepath=None):
		self._filepath = filepath

	def path2gcode(self, path, samples_per_unit=100, d=4):
		gcode = []
		if isinstance(path, str):
			path = Path(path)

		def rv(v):
			return ('%*f' % (d, round(v, d))).rstrip("0").rstrip(".")

		for segment in path:
			subdiv=max(1, round(segment.length(error=1e-5) * samples_per_unit))

			if isinstance(segment, Move):
				gcode.append('G0 X%s Y%s' % (rv(segment.end.x), rv(-segment.end.y)))
			elif isinstance(segment, (Line, Close)):
				gcode.append('G1 X%s Y%s' % (rv(segment.end.x), rv(-segment.end.y)))
			elif isinstance(segment, Arc) and abs(segment.rx - segment.ry) < 1e-9:
				# Strictly speaking, svg arcs can be non circular, whereas gcode only permits circular arcs.
				garc = "G02" if segment.sweep > 0 else "G03"
				gcode.append('%s X%s Y%s R%s' % (garc,
												 rv(segment.end.x),
												 rv(-segment.end.y),
												 rv(segment.rx)))
			else:  # Non-circular arc, Cubic or Quad Bezier Curves.
				subdiv_points = numpy.linspace(0, 1, subdiv, endpoint=True)[1:]
				points = segment.npoint(subdiv_points)  # numpy accelerated point() call
				gcode.extend(['G1 X%s Y%s' % (rv(sp[0]), rv(-sp[1])) for sp in points])

		return '\n'.join(gcode)

	def get_gcode(self, scale=1.0/96.0, samples_per_unit=100, digits=4, ppi=96.0):
		"""
		Parse gcode from an SVG file.

		scale: unit scaling between svg pixels and desired units. 1.0/96.0 is inches.
		subdivratio: How many subdivisions per unit? 1/100th inch steps.
		digits: How many digits of gcode accuracy.
		ppi: pixels per inch of the file being loaded. 96 is standard.
		"""
		gcode = []
		transform = "scale(%g)" % scale if scale != 1.0 else None
		svg = SVG.parse(self._filepath, reify=False, ppi=ppi, transform=transform)
		for element in svg.elements():
			if isinstance(element, Shape):
				if not isinstance(element, Path):
					element = Path(element)
				gcode.append({'id': element.id, 'path': self.path2gcode(element.reify(), samples_per_unit, digits)})
		return gcode




# Placed into Public Domain in June 2006 by Sean D. Spencer

# Sean D. Spencer
# sean_don4@lycos.com
# 2/19/2006
# Last Revision: 4/19/2007

# MIDI Parsing Library for Python.

TRUE = -1
FALSE = 0


class format:
    SingleTrack = 0
    MultipleTracksSync = 1
    MultipleTracksAsync = 2


class voice:
    NoteOff = 0x80
    NoteOn = 0x90
    PolyphonicKeyPressure = 0xA0 # note aftertouch
    ControllerChange = 0xB0
    ProgramChange = 0xC0
    ChannelPressure = 0xD0
    PitchBend = 0xE0


class meta:
    FileMetaEvent = 0xFF
    SMPTEOffsetMetaEvent = 0x54
    SystemExclusive = 0xF0
    SystemExclusivePacket = 0xF7
    SequenceNumber = 0x00
    TextMetaEvent = 0x01
    CopyrightMetaEvent = 0x02
    TrackName = 0x03
    InstrumentName = 0x04
    Lyric = 0x05
    Marker = 0x06
    CuePoint = 0x07
    ChannelPrefix = 0x20
    MidiPort = 0x21
    EndTrack = 0x2F
    SetTempo = 0x51
    TimeSignature = 0x58
    KeySignature = 0x59
    SequencerSpecificMetaEvent = 0x7F


class EventNote:
    def __init__(self):
        self.note_no = None
        self.velocity = None


class EventValue:
    def __init__(self):
        self.type = None
        self.value = None


class EventAmount:
    def __init__(self):
        self.amount = None


class MetaEventKeySignature:
    def __init__(self):
        self.fifths = None
        self.mode = None


class MetaEventTimeSignature:
    def __init__(self):
        self.numerator = None
        self.log_denominator = None
        self.midi_clocks = None
        self.thirty_seconds = None


class MetaEventText:
    def __init__(self):
        self.length = None
        self.text = None


class MetaEventSMPTEOffset:
    def __init__(self):
        self.hour = None
        self.minute = None
        self.second = None
        self.frame = None
        self.sub_frame = None


class MetaValues:
    def __init__(self):
        self.length = None
        self.values = None


def getNumber(theString, length):
    # MIDI uses big-endian for everything
    sum = 0
    #print "Length: " + str(length) + "  strlen: " + str(len(theString))
    for i in range(length):
        #sum = (sum *256) + int(str[i])
        sum = (sum << 8) + ord(theString[i])
    return sum, theString[length:]


def getVariableLengthNumber(str):
    sum = 0
    i = 0
    while 1:
        x = ord(str[i])
        i = i + 1
        # sum = (sum * 127) + (x (mask) 127) # mask off the 7th bit
        sum = (sum << 7) + (x & 0x7F)
        # Is 7th bit clear?
        if not (x & 0x80):
            return sum, str[i:]


def getValues(str, n=16):
    temp = []
    for x in str[:n]:
        temp.append(repr(ord(x)))
    return temp


class File:
    def __init__(self, file):
        self.file = file
        self.format = None
        self.num_tracks = None
        self.division = None
        self.tracks = []

        self.file = open(self.file, 'rb')
        str = self.file.read()
        self.file.close()

        self.read(str)

    def read(self, str):
        assert str[:4] == "MThd"
        str = str[4:]

        length, str = getNumber(str, 4)
        assert length == 6

        self.format, str = getNumber(str, 2)

        self.num_tracks, str = getNumber(str, 2)
        self.division, str = getNumber(str, 2)

        for i in range(self.num_tracks):
            track = Track(i+1)
            str = track.read(str)
            self.tracks.append(track)


class Track:
    def __init__(self, index):
        self.number = index
        self.length = None
        self.events = []

    def read(self, str):
        self.length, str = getNumber(str[4:], 4)
        track_str = str[:self.length]

        prev_absolute = 0
        prev_status = 0

        i = 0
        while track_str:
            event = Event(self.number, i+1)
            track_str = event.read(prev_absolute, prev_status, track_str)

            prev_absolute += event.delta
            prev_status = event.status
            self.events.append(event)
            i += 1

        return str[self.length:]


class Event:
    def __init__(self, track, index):
        self.number = index
        self.type = None
        self.delta = None
        self.absolute = None
        self.status = None
        self.channel = None

    def read(self, prev_time, prev_status, str):
        self.delta, str = getVariableLengthNumber(str)
        self.absolute = prev_time + self.delta

        # use running status?
        if not (ord(str[0]) & 0x80):
            # squeeze a duplication of the running status into the data string
            str = prev_status + str

        self.status = str[0]
        self.channel = ord(self.status) & 0xF

        # increment one byte, past the status
        str = str[1:]

        has_channel = has_meta = TRUE

        # handle voice events
        channel_msg = ord(self.status) & 0xF0
        if channel_msg == voice.NoteOn or \
                channel_msg == voice.NoteOff or \
                channel_msg == voice.PolyphonicKeyPressure:
            self.detail = EventNote()
            self.detail.note_no = ord(str[0])
            self.detail.velocity = ord(str[1])
            str = str[2:]

        elif channel_msg == voice.ControllerChange:
            self.detail = EventValue()
            self.detail.type = ord(str[0])
            self.detail.value = ord(str[1])
            str = str[2:]

        elif channel_msg == voice.ProgramChange or \
                channel_msg == voice.ChannelPressure:

            self.detail = EventAmount()
            self.detail.amount = ord(str[0])
            str = str[1:]

        elif channel_msg == voice.PitchBend:
            # Pitch bend uses high accuracy 14 bit unsigned integer.
            self.detail = EventAmount()
            self.detail.amount = (ord(str[0]) << 7) | ord(str[1])
            str = str[2:]

        else: has_channel = FALSE

        # handle meta events
        meta_msg = ord(self.status)
        if meta_msg == meta.FileMetaEvent:

            meta_msg = type = ord(str[0])
            length, str = getVariableLengthNumber(str[1:])

            if type == meta.SetTempo or \
                    type == meta.ChannelPrefix:

                self.detail = EventAmount()
                self.detail.tempo, str  = getNumber(str, length)

            elif type == meta.KeySignature:
                self.detail = MetaEventKeySignature()
                self.detail.fifths = ord(str[0])

                if ord(str[1]): self.detail.mode = "minor"
                else: self.detail.mode = "major"

                str = str[length:]

            elif type == meta.TimeSignature:
                self.detail = MetaEventTimeSignature()
                self.detail.numerator = ord(str[0])
                self.detail.log_denominator = ord(str[1])
                self.detail.midi_clocks = ord(str[2])
                self.detail.thirty_seconds = ord(str[3])
                str = str[length:]

            elif type == meta.TrackName or \
                    type == meta.TextMetaEvent or \
                    type == meta.Lyric or \
                    type == meta.CuePoint or \
                    type == meta.CopyrightMetaEvent:

                self.detail = MetaEventText()
                self.detail.length = length
                self.detail.text = str[:length]
                str = str[length:]

            elif type == meta.SMPTEOffsetMetaEvent:
                self.detail = MetaEventSMPTEOffset()
                self.detail.hour = ord(str[0])
                self.detail.minute = ord(str[1])
                self.detail.second = ord(str[2])
                self.detail.frame = ord(str[3])
                self.detail.sub_frame = ord(str[4])
                str = str[length:]

            elif type == meta.EndTrack:
                str = str[length:] # pass on to next track

            else: has_meta = FALSE

        elif meta_msg == meta.SystemExclusive or \
                meta_msg == meta.SystemExclusivePacket:
            self.detail = MetaValues()
            self.detail.length, str = getVariableLengthNumber(str)
            self.detail.values = getValues(str, self.detail.length)
            str = str[self.detail.length:]

        else: has_meta = FALSE

        if has_channel:
            self.type = channel_msg
        elif has_meta:
            self.type = meta_msg
        else:
            #raise "Unknown event."
            self.type = None
        return str



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	DodoLaSaumure
# Date:	30-Dec-2019

from __future__ import absolute_import
from __future__ import print_function
__author__ = "DodoLaSaumure"
__email__  = ""


import math
from bmath import Vector
from CNC import CW,CNC,Block
from ToolsPage import Plugin


#==============================================================================
# SimpleLine class
#==============================================================================
class SimpleLine:
	def __init__(self, name): 
		self.name = name

	def calc(self, xstart,ystart,xend,yend):
		points = []
		points.append(Vector(xstart,ystart))
		points.append(Vector(xend,yend))
		first = points[0]
		last = points[1]
		blocks = []
		block = Block(self.name)
		block.append(CNC.grapid(first.x(), first.y()))
		block.append(CNC.grapid(z=0.0))
		block.append("(entered)")
		block.append(CNC.gline(last.x(), last.y()))
		block.append("(exiting)")
		block.append(CNC.grapid(z=CNC.vars["safe"]))
		blocks.append(block)
		return blocks


#==============================================================================
# Create a simple Line
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Generate a simple line")

	def __init__(self, master):
		Plugin.__init__(self, master, "SimpleLine")
		self.icon  = "SimpleLine"
		self.group = "Generator"
		self.variables = [
			("xstart",        "float",    10, _("xStart")),
			("xend",        "float",    20, _("xEnd")),
			("ystart",        "float",    10, _("yStart")),
			("yend",        "float",    20, _("yEnd")),
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		n = self["name"]
		if not n or n=="default":
			n="SimpleLine"
		simpleLine = SimpleLine(n)
		blocks = simpleLine.calc(self["xstart"], self["ystart"], self["xend"], self["yend"], )
		active = app.activeBlock()
		if active==0:
			active=1
		app.gcode.insBlocks(active, blocks, _("Create Simple Line"))
		app.refresh()
		app.setStatus(_("Generated: Simple Line"))

if __name__=="__main__":
	simpleLine = SimpleLine()
	simpleLine.calc(10,10,20,20)



#!/usr/bin/python
# -*- coding: ascii -*-

# Author: @harvie Tomas Mudrunka
# Date: 7 july 2018

from __future__ import absolute_import
from __future__ import print_function
__author__ = "@harvie Tomas Mudrunka"
#__email__  = ""

__name__ = _("Center")
__version__ = "0.1"

import math
import os.path
import re
from CNC import CNC,Block
from ToolsPage import Plugin
from math import pi, sqrt, sin, cos, asin, acos, atan2, hypot, degrees, radians, copysign, fmod


class Tool(Plugin):
	__doc__ = _("""Find center of bounding box""")			#<<< This comment will be show as tooltip for the ribbon button

	def __init__(self, master):
		Plugin.__init__(self, master,"Center")
		#Helical_Descent: is the name of the plugin show in the tool ribbon button
		self.icon = "centerpoint"			#<<< This is the name of file used as icon for the ribbon button. It will be search in the "icons" subfolder
		self.group = "CAM"	#<<< This is the name of group that plugin belongs
		self.oneshot = True
		#Here we are creating the widgets presented to the user inside the plugin
		#Name, Type , Default value, Description
		self.variables = [			#<<< Define a list of components for the GUI
			("name"    ,    "db" ,    "", _("Name"))							#used to store plugin settings in the internal database
		]
		self.buttons.append("exe")  #<<< This is the button added at bottom to call the execute method below


	# ----------------------------------------------------------------------
	# This method is executed when user presses the plugin execute button
	# ----------------------------------------------------------------------
	def execute(self, app):
		#print("go!")
		blocks  = []
		for bid in app.editor.getSelectedBlocks():
			if len(app.gcode.toPath(bid)) < 1: continue
			path = app.gcode.toPath(bid)[0]
			x,y = path.center()
			eblock = Block("center of "+app.gcode[bid].name())
			eblock.append("G0 x"+str(round(x,CNC.digits))+" y"+str(round(y,CNC.digits)))
			eblock.append("G1 Z0 F200")
			eblock.append("G0 Z10")
			blocks.append(eblock)


		#active = app.activeBlock()
		#if active == 0: active+=1
		active=-1 #add to end
		app.gcode.insBlocks(active, blocks, "Center created") #<<< insert blocks over active block in the editor
		app.refresh()                                                                                           #<<< refresh editor
		app.setStatus(_("Generated: Center"))                           #<<< feed back result
		#app.gcode.blocks.append(block)



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	Filippo Rivato
# Date:	2015/10/04

from __future__ import absolute_import
from __future__ import print_function
__author__ = "Filippo Rivato"
__email__  = "f.rivato@gmail.com"

__name__ = _("Flatten")
__version__= "0.0.2"

from CNC import CNC,Block
from ToolsPage import Plugin
from Utils import to_zip

#==============================================================================
#Flatten class
#==============================================================================
class Flatten:
	def __init__(self,name="Flatten"):
		self.name = name

	#----------------------------------------------------------------------
	def make(self,app, XStart=0.0, YStart=0.0, FlatWidth=10., FlatHeight=10., \
			FlatDepth=0, BorderPass=False, CutDirection="Climb", PocketType="Raster"):

		#GCode Blocks
		blocks = []

		#Check parameters
		if CutDirection == "":
			app.setStatus(_("Flatten abort: Cut Direction is undefined"))
			return

		if PocketType == "":
			app.setStatus(_("Flatten abort: Pocket Type is undefined"))
			return

		if FlatWidth <= 0 or FlatHeight <= 0 :
			app.setStatus(_("Flatten abort: Flatten Area dimensions must be > 0"))
			return

		if FlatDepth > 0 :
			app.setStatus(_("Flatten abort: Hey this is only for subtractive machine! Check depth!"))
			return

		#Add Region disabled to show worked area
		block = Block(self.name + " Outline")
		block.enable = False
		block.append(CNC.zsafe())
		xR,yR = self.RectPath(XStart,YStart,FlatWidth,FlatHeight)
		for x,y in zip(xR,yR):
			block.append(CNC.gline(x,y))
		blocks.append(block)

		# Load tool and material settings
		toolDiam = CNC.vars['diameter']
		toolRadius = toolDiam / 2.

		#Calc tool diameter with Maximum Step Over allowed
		StepOverInUnitMax = toolDiam * CNC.vars['stepover'] / 100.0

		#Offset for Border Cut
		BorderXStart = XStart + toolRadius
		BorderYStart = YStart + toolRadius
		BorderWidth = FlatWidth - toolDiam
		BorderHeight = FlatHeight - toolDiam
		BorderXEnd = XStart + FlatWidth - toolRadius
		BorderYEnd = YStart + FlatHeight - toolRadius

		PocketXStart = BorderXStart
		PocketYStart = BorderYStart
		PocketXEnd = BorderXEnd
		PocketYEnd = BorderYEnd

		#Calc space to work with/without border cut
		WToWork = FlatWidth - toolDiam
		HToWork = FlatHeight - toolDiam

		if(WToWork < toolRadius or HToWork < toolRadius):
			app.setStatus(_("Flatten abort: Flatten area is too small for this End Mill."))
			return

		#Prepare points for pocketing
		xP=[]
		yP=[]
        #and border
		xB=[]
		yB=[]

        #---------------------------------------------------------------------
        #Raster approach
		if PocketType == "Raster":
			#Correct sizes if border is used
			if(BorderPass):
				PocketXStart += StepOverInUnitMax
				PocketYStart += StepOverInUnitMax
				PocketXEnd -= StepOverInUnitMax
				PocketYEnd -= StepOverInUnitMax
				WToWork -= (StepOverInUnitMax)
				HToWork -= (StepOverInUnitMax)

			#Calc number of pass
			VerticalCount = (int)(HToWork / StepOverInUnitMax)
			#Calc step minor of Max step
			StepOverInUnit = HToWork / (VerticalCount +1)
			flip = False
			ActualY = PocketYStart
			#Zig zag
			if StepOverInUnit==0 : StepOverInUnit=0.001  #avoid infinite while loop
			while (True):
				#Zig
				xP.append(self.ZigZag(flip,PocketXStart,PocketXEnd))
				yP.append(ActualY)
				flip = not flip
				#Zag
				xP.append(self.ZigZag(flip,PocketXStart,PocketXEnd))
				yP.append(ActualY)
				if(ActualY >= PocketYEnd - StepOverInUnitMax + StepOverInUnit):
					break
				#Up
				ActualY += StepOverInUnit
				xP.append(self.ZigZag(flip,PocketXStart,PocketXEnd))
				yP.append(ActualY)

			#Points for border cut depends on Zig/Zag end
			if(BorderPass):
				if flip:
					xB,yB = self.RectPath(BorderXStart,BorderYEnd,BorderWidth,-BorderHeight)
				else:
					xB,yB = self.RectPath(BorderXEnd,BorderYEnd,-BorderWidth,-BorderHeight)

				#Reverse in case of Climb
				if CutDirection == "Climb":
					xB = xB[::-1]
					yB = yB[::-1]

		#---------------------------------------------------------------------
        #Offset approach
		if PocketType == "Offset":
			#Calc number of pass
			VerticalCount = (int)(HToWork / StepOverInUnitMax)
			HorrizontalCount = (int)(WToWork / StepOverInUnitMax)
			#Make them odd
			if VerticalCount%2 == 0 : VerticalCount += 1
			if HorrizontalCount%2 == 0 : HorrizontalCount += 1
			#Calc step minor of Max step
			StepOverInUnitH = HToWork / (VerticalCount)
			StepOverInUnitW = WToWork / (HorrizontalCount)

			#Start from border to center
			xS = PocketXStart
			yS = PocketYStart
			wS = WToWork
			hS = HToWork
			xC = 0
			yC = 0
			while (xC<=HorrizontalCount/2 and yC<=VerticalCount/2):
				#Pocket offset points
				xO,yO = self.RectPath(xS, yS, wS, hS)
				if CutDirection == "Conventional":
					xO = xO[::-1]
					yO = yO[::-1]

				xP = xP + xO
				yP = yP + yO
				xS+=StepOverInUnitH
				yS+=StepOverInUnitW
				hS-=2.0*StepOverInUnitH
				wS-=2.0*StepOverInUnitW
				xC += 1
				yC += 1

			#Reverse point to start from inside (less stress on the tool)
			xP = xP[::-1]
			yP = yP[::-1]

		#Blocks for pocketing
		block = Block(self.name)
		block.append("(Flatten from X=%g Y=%g)"%(XStart,YStart))
		block.append("(W=%g x H=%g x D=%g)"%(FlatWidth,FlatHeight,FlatDepth))
		block.append("(Approach: %s %s)" % (PocketType,CutDirection))
		if BorderPass : block.append("(with border)")

		#Move safe to first point
		block.append(CNC.zsafe())
		block.append(CNC.grapid(xP[0],yP[0]))
		#Init Depth
		currDepth = 0.
		stepz = CNC.vars['stepz']
		if stepz==0 : stepz=0.001  #avoid infinite while loop

		#Create GCode from points
		while True:
			currDepth -= stepz
			if currDepth < FlatDepth : currDepth = FlatDepth
			block.append(CNC.zenter(currDepth))
			block.append(CNC.gcode(1, [("f",CNC.vars["cutfeed"])]))

			#Pocketing
			lastxy = None
			for x,y in zip(xP,yP):
#				block.append(CNC.gline(x,y))
				if lastxy != CNC.gline(x,y) or None:
					block.append(CNC.gline(x,y))
				lastxy = CNC.gline(x,y)

			#Border cut if request
			for x,y in zip(xB,yB):
				block.append(CNC.gline(x,y))

			#Verify exit condition
			if currDepth <= FlatDepth : break

			#Move to the begin in a safe way
			block.append(CNC.zsafe())
			block.append(CNC.grapid(xP[0],yP[0]))

		#Zsafe
		block.append(CNC.zsafe())
		blocks.append(block)
		return blocks

	#----------------------------------------------------------------------
	def RectPath(self,x,y,w,h):
		xR = []
		yR = []
		xR.append(x)
		yR.append(y)
		xR.append(x + w)
		yR.append(y)
		xR.append(x + w)
		yR.append(y + h)
		xR.append(x)
		yR.append(y + h)
		xR.append(x)
		yR.append(y)
		return (xR,yR)

	#----------------------------------------------------------------------
	def ZigZag(self,flip,zig,zag):
		if flip:
			return zig
		else:
			return zag


#==============================================================================
# Create a flatten surface
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Flatten an area in different ways")

	def __init__(self, master):
		Plugin.__init__(self, master, "Flatten")
		self.icon  = "flatten"
		self.group = "CAM"
		self.variables = [
			("name",           "db",    "", _("Name")),
			("XStart"  ,       "mm",   0.0, _("X start")),
			("YStart"  ,       "mm",   0.0, _("Y start")),
			("FlatWidth" ,     "mm",   30.0, _("Width to flatten")),
			("FlatHeight"  ,   "mm",   20.0, _("Height to flatten")),
			("FlatDepth"  ,    "mm",    0.0, _("Depth to flatten")),
			("BorderPass"  , "bool",  True , _("Raster border")),
			("CutDirection", "Climb,Conventional","Climb", _("Cut Direction")),
			("PocketType"  , "Raster,Offset" ,"Raster", _("Pocket type"))
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		n = self["name"]
		if not n or n=="default": n="Flatten"
		flatten = Flatten(n)

		blocks = flatten.make(app,
				self.fromMm("XStart"),
				self.fromMm("YStart"),
				self.fromMm("FlatWidth"),
				self.fromMm("FlatHeight"),
				self.fromMm("FlatDepth"),
				self["BorderPass"],
				self["CutDirection"],
				self["PocketType"]
				)

		if blocks is not None:
			active = app.activeBlock()
			if active==0: active=1
			app.gcode.insBlocks(active, blocks, "Flatten")
			app.refresh()
			app.setStatus(_("Flatten: Generated flatten surface"))



import shutil
import time

import pyautogui

from .base import BaseGUITestCase


class SmokeTest(BaseGUITestCase):
    def test_launches(self):
        self.assertFalse(self.gui_proc.poll())

    def test_can_load_and_run_sample_gcode(self):
        shutil.copy(
            self.get_static_path('sample.gcode'),
            '/tmp/',
        )

        self.send_command('load /tmp/sample.gcode')
        pyautogui.press('f10')  # Mapped to 'start' in config
        self.delay(10)
        pyautogui.press('f12')  # Mapped to 'stop' in config
        self.delay(10)

        state = self.get_bcnc_state()
        # Assert that we moved the tool position
        self.assertGreaterEqual(state['wx'], 0)
        self.assertGreaterEqual(state['wy'], 0)



#!/usr/bin/python
# -*- coding: ascii -*-

# Author: @DodoLaSaumure Pierre KLein
# Date: 9 feb 2021

from __future__ import print_function
from __future__ import print_function
__author__ = "@DodoLaSaumure  (Pierre Klein)"
#__email__  = ""

__name__ = _("PocketIsland")
__version__ = "0.0.1"

import math
import os.path
import re
from CNC import CNC,Block#,toPath,importPath,addUndo
from ToolsPage import Plugin
from math import pi, sqrt, sin, cos, asin, acos, atan2, hypot, degrees, radians, copysign, fmod
from bpath import EPS,eq,Path, Segment
from bmath import Vector
from copy import deepcopy

try:
	import Tkinter
	from Tkinter import *
	import tkMessageBox
except ImportError:
	import tkinter
	from tkinter import *
	import tkinter.messagebox as tkMessageBox

# =============================================================================
#==============================================================================

def pocket(blocks, RecursiveDepth,ProfileDir,CutDir,AdditionalCut,Overcuts, CustomRecursiveDepth,
		allowG1, diameter, stepover, name,gcode,items):
	undoinfo = []
	msg = ""
	newblocks = []
	islandslist = []
	for bid,block in enumerate(gcode.blocks):
		if block.operationTest('island'):
			for islandPath in gcode.toPath(bid):
				islandslist.append(islandPath)
	for bid in reversed(blocks):
		if gcode.blocks[bid].name() in ("Header", "Footer"): continue
		newpath = []
		for path in gcode.toPath(bid):
			if not path.isClosed():
				m = "Path: '%s' is OPEN"%(path.name)
				if m not in msg:
					if msg: msg += "\n"
					msg += m
				path.close()

			# Remove tiny segments
			path.removeZeroLength(abs(diameter)/100.)
			# Convert very small arcs to lines
			path.convert2Lines(abs(diameter)/10.)

			remove = ["cut","reverse","climb","conventional","cw","ccw","pocket"]
			if name is None:
				path.name = Block.operationName(path.name, "pocket,conventional,cw", remove)
			else:
				path.name = Block.operationName(path.name, name, remove)
			MyPocket = PocketIsland([path],RecursiveDepth,ProfileDir,CutDir,AdditionalCut,
								Overcuts,CustomRecursiveDepth,
								allowG1,diameter,stepover,0,islandslist)
			newpathList =  MyPocket.getfullpath()
			#concatenate newpath in a single list and split2contours
			if allowG1 :
				MyFullPath = Path("Pocket")
				for path in newpathList :
					for seg in path:
						MyFullPath.append(seg)
				newpathList = MyFullPath.split2contours()
		if newpathList:
			# remember length to shift all new blocks
			# the are inserted before
			before = len(newblocks)
			undoinfo.extend(gcode.importPath(bid+1, newpathList,
				newblocks, True, False))
			new = len(newblocks)-before
			for i in range(before):
				newblocks[i] += new
			gcode.blocks[bid].enable = False
	gcode.addUndo(undoinfo)

	# return new blocks inside the blocks list
	del blocks[:]
	blocks.extend(newblocks)
	return msg

class PocketIsland:
	def __init__(self,pathlist,RecursiveDepth,ProfileDir,CutDir,AdditionalCut, Overcuts,CustomRecursiveDepth,
				allowG1,diameter,stepover,depth,islandslist=[]):
		self.outpaths = pathlist
		self.islands = islandslist
		self.diameter = diameter
		self.stepover = stepover
		self.RecursiveDepth=RecursiveDepth
		self.ProfileDir=ProfileDir
		self.CutDir=CutDir
		self.AdditionalCut=float(AdditionalCut)
		self.Overcuts = bool(Overcuts)
		self.CustomRecursiveDepth=CustomRecursiveDepth
		self.childrenIslands = []
		self.childrenOutpath = []
		self.fullpath = []
		self.depth = depth
		self.islandG1SegList = Path("islandG1SegList")
		self.outPathG1SegList = Path("outPathG1SegList")
		self.allowG1 = allowG1
		maxdepthchoice = {"Single profile":0,
						"Custom recursive depth":int(self.CustomRecursiveDepth-1),
						"Full pocket":100}
		profileDirChoice = {"inside":1.,"outside":-1.}
		cutDirChoice = {"conventional milling":1.,"climbing milling":-1.}
		self.selectCutDir = cutDirChoice.get(self.CutDir,1.)
		self.profiledir = profileDirChoice.get(self.ProfileDir,1.)
		if self.RecursiveDepth=="Full pocket" :
			self.profiledir=1.#to avoid making full pockets, with full recursive depth, outside the path
		maxdepth=maxdepthchoice.get(self.RecursiveDepth,0)
		import sys
		sys.setrecursionlimit(max(sys.getrecursionlimit(),maxdepth+10))
		if depth>maxdepth: return None
		self.eliminateOutsideIslands()
		self.inoutprofile()
		self.removeOutofProfileLinkingSegs()
		self.interesect()
		self.removeOutOfProfile()
		self.removeInsideIslands()
		self.getNewPathAndIslands()
		self.getPaths()
		if len (self.CleanPath)>0:
			self.recurse()

	def eliminateOutsideIslands(self):
		self.insideIslandList = []
		for island in self.islands:
			for path in self.outpaths :
				if island.isPathInside(path)>=0:
					self.insideIslandList.append(island)

	def inoutprofile(self):
		if self.depth == 0:
			self.offset = -self.diameter / 2.0 +self.AdditionalCut
		else:
			self.offset = -self.diameter*self.stepover
		self.OutOffsetPathList = []
		for path in self.outpaths :
			p1=p2=None
			if len(path)>0:
				p1 = path[0].A
			if self.depth == 0 :
				path.directionSet(self.selectCutDir*float(self.profiledir))
			direct = path.direction()
			opath = path.offset(self.profiledir*self.offset*float(direct))
			opath.intersectSelf()
			opath.removeExcluded(path, abs(self.offset))
			if self.depth == 0 and self.Overcuts:
				opath.overcut(self.profiledir*self.offset*float(direct))
			if len(opath)>0:
				p2 = opath[0].A
				self.OutOffsetPathList.append(opath)
				if self.depth >0  and p1 is not None:
					self.outPathG1SegList.append(Segment(Segment.LINE,p1,p2))
		self.islandOffPaths = []
		for island in self.insideIslandList :
			p3=p4=None
			if len(island)>0:
				p3 = island[0].A
			if self.depth == 0 :
				island.directionSet(-self.selectCutDir*float(self.profiledir))
			direct = island.direction()
			offIsl = island.offset(-self.profiledir*self.offset*float(direct))
			if len(offIsl)>0:
				p4 = offIsl[0].A
			if self.depth >0 and p3 is not None and p4 is not None :
				self.islandG1SegList.append(Segment(Segment.LINE,p3,p4))
			offIsl.intersectSelf()
			offIsl.removeExcluded(island, abs(self.offset))
			if self.depth == 0 and self.Overcuts:
				offIsl.overcut(-self.profiledir*self.offset*float(direct))
			self.islandOffPaths.append(offIsl)


	def removeOutofProfileLinkingSegs(self):
		self.tmpoutG1 = deepcopy(self.outPathG1SegList)
		self.tmpinG1 = deepcopy(self.islandG1SegList)
		for i,seg in enumerate(self.outPathG1SegList) :
			for path in self.islandOffPaths :
				inside =path.isSegInside(seg)==1#outseg inside offsetislands =>pop
				if  inside and seg in self.tmpoutG1:
					self.tmpoutG1.remove(seg)
		for i,seg in enumerate(self.islandG1SegList):
			for path in self.OutOffsetPathList :
				outside = path.isSegInside(seg)<1#inseg outside offsetOutpaths => pop
				if outside and  seg in self.tmpinG1 : 
						self.tmpinG1.remove(seg)
		self.outPathG1SegList = self.tmpoutG1
		self.islandG1SegList = self.tmpinG1

	def interesect(self):
		self.IntersectedIslands = []
		for island in self.islandOffPaths :
			for path in self.OutOffsetPathList :
				path.intersectPath(island)
				island.intersectPath(path)
			for island2 in self.islandOffPaths :
				island.intersectPath(island2)
			self.IntersectedIslands.append(island)

	def removeOutOfProfile(self):
		self.NewPaths = []
		newoutpath = Path("path")
		for path in self.OutOffsetPathList: 
			for seg in path:
				newoutpath.append(seg)
		for OutoffsetPath in self.OutOffsetPathList:
			for path in self.IntersectedIslands :
				for seg in path :
					inside = not OutoffsetPath.isSegInside(seg)==-1
					if inside:
						newoutpath.append(seg)
		purgednewoutpath= newoutpath.split2contours()#list of paths
		self.NewPaths.extend(purgednewoutpath)

	def removeInsideIslands(self):
		self.CleanPath = []
		cleanpath = Path("Path")
		for path in self.NewPaths :
			for seg in path :
				inside = False
				for island in self.IntersectedIslands :
					issegin = island.isSegInside(seg)==1
					if issegin:
						if not seg in island:
							inside = True
							break
				if not inside:
					cleanpath.append(seg)
		cleanpath = cleanpath.split2contours()
		self.CleanPath.extend(cleanpath)

	def getNewPathAndIslands(self):
		if len(self.CleanPath)==1:
			self.childrenOutpath = self.CleanPath
		else :
			for elt in self.CleanPath: #List of paths
				for elt2 in self.CleanPath :
						ins = elt2.isPathInside(elt)==1
						ident = elt2.isidentical(elt)
						addedelt2  = elt2 in self.childrenIslands 
						if ins and not ident and not addedelt2 :
							self.childrenIslands.append(elt2)
			for elt in self.CleanPath: #List of paths
				for elt2 in self.CleanPath:
						if not elt2 in self.childrenIslands and not elt2 in self.childrenOutpath:
							self.childrenOutpath.append(elt2)

	def getPaths(self):
		if len (self.CleanPath)>0:
			if len(self.islandG1SegList) >0 :
				self.outPathG1SegList.extend(self.islandG1SegList)
			if self.allowG1 and len(self.outPathG1SegList)>0:
				for seg in self.outPathG1SegList :
					path = Path("SegPath")
					path.append(seg)
					self.CleanPath.append(path)
			self.fullpath.extend(self.CleanPath)
		return self.CleanPath

	def recurse(self):
		pcket = PocketIsland(self.childrenOutpath,self.RecursiveDepth,self.ProfileDir,
							self.CutDir,self.AdditionalCut,self.Overcuts, self.CustomRecursiveDepth,
							self.allowG1,self.diameter,self.stepover,self.depth+1,self.childrenIslands)
		self.fullpath.extend(pcket.getfullpath())

	def getfullpath(self) :
		return self.fullpath


class Tool(Plugin):
	__doc__ = _("Generate a pocket with Inside Islands")

	def __init__(self, master):
		Plugin.__init__(self, master, "PocketIsland")
		self.icon  = "pocketisland"
		self.group = "Development"
		self.variables = [
			("name",      "db" ,    "", _("Name")),
			("endmill",   "db" ,    "", _("End Mill")),
			("RecursiveDepth","Single profile,Full pocket,Custom recursive depth", "Single profile",  _("Recursive depth")),
			("ProfileDir","inside,outside", "inside",  _("Profile direction if profile option selected")),
			("CutDir","conventional milling,climbing milling", "conventional milling",  _("Cut Direction,default is conventional")),
			("AdditionalCut"  ,         "mm" ,     0., _("Additional cut inside profile")),
			("Overcuts"  ,         "bool" ,     False, _("Overcuts inside corners")),
			("CustomRecursiveDepth","int",1,_("Nb of contours (Custom Recursive Depth)")),
			("allowG1",        "bool",    True, _("allow pocket linking segments(default yes)")),
		]
		self.buttons.append("exe")
	# ----------------------------------------------------------------------
	def execute(self, app):
		if self["endmill"]:
			self.master["endmill"].makeCurrent(self["endmill"])
		RecursiveDepth=self["RecursiveDepth"]
		ProfileDir=self["ProfileDir"]
		CutDir=self["CutDir"]
		AdditionalCut=self["AdditionalCut"]
		Overcuts = self["Overcuts"]
		CustomRecursiveDepth=self["CustomRecursiveDepth"]
		allowG1 = self["allowG1"]
		name = self["name"]
		if name=="default" or name=="": name=None
		tool = app.tools["EndMill"]
		diameter = app.tools.fromMm(tool["diameter"])
		try:
			stepover = tool["stepover"] / 100.0
		except TypeError:
			stepover = 0.

		app.busy()
		blocks = app.editor.getSelectedBlocks()
 
		msg = pocket(blocks,RecursiveDepth,ProfileDir,CutDir,
					AdditionalCut, Overcuts,CustomRecursiveDepth,
					bool(allowG1), diameter, stepover,
					name,gcode = app.gcode,items=app.editor.getCleanSelection())
		if msg:
			tkMessageBox.showwarning(_("Open paths"),
					_("WARNING: %s")%(msg),
					parent=app)
		app.editor.fill()
		app.editor.selectBlocks(blocks)
		app.draw()
		app.notBusy()
		app.setStatus(_("Generate pocket path"))

##############################################



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	Filippo Rivato
# Date:     18 November 2015
# Porting of image2gcode and based on dmap2gcode
# Source was used from the following works:
#              image-to-gcode.py   2005 Chris Radek chris@timeguy.com
#              image-to-gcode.py   2006 Jeff Epler
#              Author.py(linuxcnc) 2007 Jeff Epler  jepler@unpythonic.net
#              dmap2gcode G-Code Generator 2015  @schorchworks

from __future__ import absolute_import
from __future__ import print_function
__author__ = "Filippo Rivato"
__email__  = "f.rivato@gmail.com"

__name__ = _("Heightmap")
__version__= "0.0.1"

import math
from CNC import CNC,Block
from ToolsPage import Plugin

from imageToGcode import *


#==============================================================================
#Heightmap class
#==============================================================================
class Heightmap:
	def __init__(self,name="Heightmap"):
		self.name = name


#==============================================================================
# Create heightmap
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Use a brightess map to create a variable Z path")

	def __init__(self, master):
		Plugin.__init__(self, master, "Heightmap")
		self.icon  = "heightmap"
		self.group = "Generator"

		self.variables = [
			("name",      "db" ,     "", _("Name")),
			("Depth",     "mm" ,   -1.0, _("Working Depth")),
			("MaxSize",   "mm" ,  100.0, _("Maximum size")),
			("Scan",      "Columns,Rows,C&R,R&C", "Rows", _("Scan")),
			("ScanDir",   "Alternating,Positive,Negative,Up Mill,Down Mill" ,\
						"Alternating" , _("ScanDir")),
			("CutTop",    "bool",  False, _("Cut Top")),
			("CutBorder", "bool",  False, _("Cut Border")),
			("Invert",    "bool",  False, _("Invert")),
			("SinglePass",    "bool",  False, _("Single pass")),
			("File",      "file" ,	  "", _("Image to process")),
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		#Try import PIL
		try:
			from PIL import Image
		except:
			app.setStatus(_("Heightmap abort: This plugin requires PIL/Pillow"))
			return

		#Try read image
		fileName = self["File"]
		try:
			img = Image.open(fileName)
			img = img.convert ('L') #Luminance
		except:
			app.setStatus(_("Heightmap abort: Can't read image file"))
			return

		if self.fromMm("Depth")>=0:
			app.setStatus(_("Heightmap abort: depth must be < 0"))
			return

		#Define type of matrix manipulation
		NUMPY = True
		if NUMPY == True:
			try:
				import numpy
			except:
				NUMPY = False

		if NUMPY == True:
			Image_Matrix = Image_Matrix_Numpy
		else:
			Image_Matrix = Image_Matrix_List
			#print "Install NumPy will speed up heightmap creation"

		MAT = Image_Matrix()
		MAT.FromImage(img,True)

		maxSize = self.fromMm("MaxSize")
		w, h = img.size

		if (w > h):
			ratio = float(w) / float(h)
			image_w = maxSize
			image_h = maxSize / ratio
		else:
			ratio = float(h) / float(w)
			image_w = maxSize / ratio
			image_h = maxSize

		#Calc pixel size
		pixel_size = image_h / ( float(MAT.width) - 1.0 )

		tolerance     =  0.1
		safe_z        =  CNC.vars["safe"]
		splitstep     =  0.0	#Offset Stepover
		toptol        = -0.1	#Top Tolerance
		depth         = -self["Depth"]
		Cont_Angle    =  45.0 #Contact angle , only with "Lace Bounding"

		#Cut perimeter/border
		cutperim      = 0
		if (self["CutBorder"]) : cutperim = 1

		######################################################
		tool = app.tools["EndMill"]
		tool_shape = tool["shape"]

		tool_diameter =  CNC.vars['diameter']
		feed_rate     =  CNC.vars["cutfeed"]

		zStep         =  CNC.vars['stepz']
		if self["SinglePass"]:
			zStep     =  0.0
		rough_offset  =  0.0
		rough_feed    =  CNC.vars["cutfeed"]

		plunge_feed   =  CNC.vars["cutfeedz"]
		stepover      =  tool_diameter * CNC.vars['stepover'] / 100.0
		step          =  max(1, int(math.floor( float(stepover) / pixel_size)))

		edge_offset   = 0
		######################################################
		if tool_shape == "Square End" or tool_shape == "Fishtail" or tool_shape == "Radiused":
			TOOL = make_tool_shape(NUMPY,endmill, tool_diameter, pixel_size)
		elif tool_shape== "V-cutting":
			try:
				v_angle = float(tool["angle"])
			except:
				app.setStatus(_("Heightmap abort: angle not defined for selected End Mill"))
				return
			TOOL = make_tool_shape(NUMPY,vee_common(v_angle), tool_diameter, pixel_size)
		else: #"Ball End"
			TOOL = make_tool_shape(NUMPY,ball_tool, tool_diameter, pixel_size)

		######################################################
		rows = 0
		columns = 0
		columns_first = 0
		scanpat = self["Scan"]
		if scanpat!= "Columns":
			rows = 1
		if scanpat!= "Rows":
			columns = 1
		if scanpat == "C&R":
			columns_first = 1

		######################################################
		# Options are "Alternating", "Positive"   , "Negative",  "Up Mill", "Down Mill"
		converter = self["ScanDir"]

		if converter == "Positive":
			conv_index = 0

		elif converter == "Negative":
			conv_index = 1

		elif converter == "Alternating":
			conv_index = 2

		elif converter == "Up Mill":
			conv_index = 3

		elif converter == "Down Mill":
			conv_index = 4
		else:
			conv_index = 2

		######################################################
		convert_makers = [ Convert_Scan_Increasing, Convert_Scan_Decreasing, \
		 Convert_Scan_Alternating, Convert_Scan_Upmill, Convert_Scan_Downmill ]
		if rows:
			convert_rows = convert_makers[conv_index]()
		else:
			convert_rows = None

		if columns:
			convert_cols = convert_makers[conv_index]()
		else:
			convert_cols = None

		######################################################
		lace_bound_val = "None" #"None","Secondary","Full"
		if lace_bound_val != "None" and rows and columns:
#			slope = tan( Cont_Angle*pi/180 )
			slope = math.tan(Cont_Angle*math.pi/180)
			if columns_first:
				convert_rows = Reduce_Scan_Lace(convert_rows, slope, step+1)
			else:
				convert_cols = Reduce_Scan_Lace(convert_cols, slope, step+1)
			if lace_bound_val == "Full":
				if columns_first:
					convert_cols = Reduce_Scan_Lace(convert_cols, slope, step+1)
				else:
					convert_rows = Reduce_Scan_Lace(convert_rows, slope, step+1)

		######################################################
		###              START COMMON STUFF                ###
		######################################################
##		units = "mm"
##		if units == "in":
##			units = 'G20'
##		else:
##			units = 'G21'
		#Units not used
		units = ""

		######################################################
		cuttop = 1
		if (self["CutTop"]) : cuttop = 0
		if cuttop:
			if rows == 1:
				convert_rows = Reduce_Scan_Lace_new(convert_rows, toptol, 1)
			if columns == 1:
				convert_cols = Reduce_Scan_Lace_new(convert_cols, toptol, 1)

		######################################################
		#Force disable arcs
		disable_arcs = True #grbl doesn't like this, G91.1?
		if (not disable_arcs):
			Entry_cut   = ArcEntryCut(plunge_feed, .125)
		else:
			Entry_cut   = SimpleEntryCut(plunge_feed)

		######################################################
		#Force normalize
		normalize = True
		if normalize:
			pass
			a = MAT.min()
			b = MAT.max()
			if a != b:
				MAT.minus(a)
				MAT.mult(1./(b-a))
		else:
			MAT.mult(1/255.0)

		xoffset = 0
		yoffset = 0
		######################################################
		MAT.mult(depth)

		##########################################
		#         ORIGIN LOCATING STUFF          #
		##########################################
		minx = 0
		maxx = image_w
		miny = 0
		maxy = image_h
		midx = (minx + maxx)/2
		midy = (miny + maxy)/2

		#Force origin, we can move it later
		origin = "Bot-Left"
		CASE = str(origin)
		if     CASE == "Top-Left":
			x_zero = minx
			y_zero = maxy
		elif   CASE == "Top-Center":
			x_zero = midx
			y_zero = maxy
		elif   CASE == "Top-Right":
			x_zero = maxx
			y_zero = maxy
		elif   CASE == "Mid-Left":
			x_zero = minx
			y_zero = midy
		elif   CASE == "Mid-Center":
			x_zero = midx
			y_zero = midy
		elif   CASE == "Mid-Right":
			x_zero = maxx
			y_zero = midy
		elif   CASE == "Bot-Left":
			x_zero = minx
			y_zero = miny
		elif   CASE == "Bot-Center":
			x_zero = midx
			y_zero = miny
		elif   CASE == "Bot-Right":
			x_zero = maxx
			y_zero = miny
		elif   CASE == "Arc-Center":
			x_zero = 0
			y_zero = 0
		else:          #"Default"
			x_zero = 0
			y_zero = 0

		xoffset = xoffset - x_zero
		yoffset = yoffset - y_zero

		######################################################
		invert = self["Invert"]
		if invert:
			MAT.mult(-1.0)
		else:
			MAT.minus(depth)

		######################################################

		gcode = []
		MAT.pad_w_zeros(TOOL)

		header = ""
		postscript = ""
		gcode = convert(self,          \
						 MAT,           \
						 units,         \
						 TOOL,          \
						 pixel_size,    \
						 step,          \
						 safe_z,        \
						 tolerance,     \
						 feed_rate,     \
						 convert_rows,  \
						 convert_cols,  \
						 columns_first, \
						 cutperim,      \
						 Entry_cut,     \
						 zStep,         \
						 rough_feed,    \
						 xoffset,       \
						 yoffset,       \
						 splitstep,     \
						 header,        \
						 postscript,    \
						 edge_offset,   \
						 disable_arcs)

		#Gcode
		n = self["name"]
		if not n or n=="default": n="Heightmap"
		block = Block(n)
		block.append("(Size: %d x %d x %d)"%(image_w,image_h,depth))
		block.append("(Endmill shape: %s , Diameter: %.3f)"%(tool_shape,tool_diameter))
		for line in gcode:
			block.append(line)

		blocks = []
		blocks.append(block)
		active = app.activeBlock()
		if active==0: active=1
		app.gcode.insBlocks(active, blocks, n)
		app.refresh()
		app.setStatus(_("Generated Heightmap %d x %d x %d ")%(image_w,image_h,depth) )



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	Filippo Rivato
# Date: December 2015

from __future__ import absolute_import
from __future__ import print_function
__author__ = "Filippo Rivato"
__email__ = "f.rivato@gmail.com"

__name__ = "Text"
__version__ = "0.0.1"

from CNC import CNC,Block
from ToolsPage import Plugin
from Utils import to_decode


#==============================================================================
#Text class
#==============================================================================
class Text:
	def __init__(self,name="Text"):
		self.name = name


#==============================================================================
# Create Text
#==============================================================================
class Tool(Plugin):
	__doc__ =  _("Create text using a ttf font")

	def __init__(self, master):
		Plugin.__init__(self, master, "Text")
		self.icon  = "text"
		self.group = "Generator"

		self.variables = [("name",      "db" ,    "", _("Name")),
				("Text",        "text", "Write this!", _("Text to generate")),
				("Depth",       "mm",    0.0, _("Working Depth")),
				("FontSize",    "mm",   10.0, _("Font size")),
				("FontFile",    "file",   "", _("Font file")),
                                ("Closed",      "bool", True, _("Close Contours")),
				("ImageToAscii","file",   "", _("Image to Ascii")),
				("CharsWidth",  "int",    80, _("Image chars width"))]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):

		#Get inputs
		fontSize      = self.fromMm("FontSize")
		depth         = self.fromMm("Depth")
		textToWrite   = to_decode(self["Text"])
		fontFileName  = self["FontFile"]
		closed        = self["Closed"]
		imageFileName = self["ImageToAscii"]
		charsWidth    = self["CharsWidth"]

		#Check parameters!!!
		if fontSize <=0:
			app.setStatus(_("Text abort: please input a Font size > 0"))
			return
		if fontFileName == "":
			app.setStatus(_("Text abort: please select a font file"))
			return
		if imageFileName != "":
			try:
				textToWrite = self.asciiArt(imageFileName,charsWidth)
			except:
				pass
		if textToWrite == "":
			textToWrite = "Nel mezzo del cammin di nostra vita..."
			return

		#Init blocks
		blocks = []
		n = self["name"]
		if not n or n == "default": n = "Text"
		block = Block(n)
		if(u'\n' in  textToWrite):
			block.append("(Text:)")
			for line in textToWrite.splitlines():
				block.append("(%s)" % line)
		else:
			block.append("(Text: %s)" % textToWrite)
		try:
			import ttf
			font = ttf.TruetypeInfo(fontFileName)
		except:
			app.setStatus(_("Text abort: That's embarrassing, I can't read this font file!"))
			return
		cmap = font.get_character_map()

		kern = None
		try:
			kern = font.get_glyph_kernings()
		except:
			pass
		adv = font.get_glyph_advances()

		xOffset = 0
		yOffset = 0
		glyphIndxLast = cmap[' ']
		for c in textToWrite:
			#New line
			if c == u'\n':
				xOffset = 0.0
				yOffset -= 1#offset for new line
				continue

			if c in cmap:
				glyphIndx = cmap[c]

				if (kern and (glyphIndx,glyphIndxLast) in kern):
					k = kern[(glyphIndx,glyphIndxLast)] #FIXME: use kern for offset??

				#Get glyph contours as line segments and draw them
				gc = font.get_glyph_contours(glyphIndx,closed)
				if(not gc):
					gc = font.get_glyph_contours(0,closed)#standard glyph for missing glyphs (complex glyph)
				if(gc and not c==' '): #FIXME: for some reason space is not mapped correctly!!!
					self.writeGlyphContour(block, font, gc, fontSize, depth, xOffset, yOffset)

				if glyphIndx < len(adv):
					xOffset += adv[glyphIndx]
				else:
					xOffset += 1
				glyphIndxLast = glyphIndx

		#Remeber to close Font
		font.close()

		#Gcode Zsafe
		block.append(CNC.zsafe())

		blocks.append(block)
		active = app.activeBlock()
		if active==0: active=1
		app.gcode.insBlocks(active, blocks, "Text")
		app.refresh()
		app.setStatus("Generated Text")


	#Write GCode from glyph contours
	def writeGlyphContour(self,block,font,contours,fontSize,depth,xO, yO):
		width = font.header.x_max - font.header.x_min
		height = font.header.y_max - font.header.y_min
		scale = fontSize / font.header.units_per_em
		xO = xO * fontSize
		yO = yO * fontSize
		for cont in contours:
			block.append("( ---------- cut-here ---------- )")
			block.append(CNC.zsafe())
			block.append(CNC.grapid(xO + cont[0].x * scale , yO + cont[0].y * scale))
			block.append(CNC.zenter(depth))
			block.append(CNC.gcode(1, [("f",CNC.vars["cutfeed"])]))
			for p in cont:
				block.append(CNC.gline(xO + p.x * scale, yO + p.y * scale))


	def image_to_ascii(self,image):
		ascii_chars = [ '#', 'A', '@', '%', 'S', '+', '<', '*', ':', ',', '.']
		image_as_ascii = []
		all_pixels = list(image.getdata())
		for pixel_value in all_pixels:
			index = pixel_value / 25 # 0 - 10
			image_as_ascii.append(ascii_chars[index])
		return image_as_ascii

	def asciiArt(self,filePath,new_width = 80):
		from PIL import Image
		img = Image.open(filePath)
		width, height = img.size
		new_height = int((height * new_width) / width)
		new_image = img.resize((new_width, new_height))
		new_image = new_image.convert("L") # convert to grayscale

		# now that we have a grayscale image with some fixed width we have to convert every pixel
		# to the appropriate ascii character from "ascii_chars"
		img_as_ascii = self.image_to_ascii(new_image)
		img_as_ascii = ''.join(ch for ch in img_as_ascii)
		output = ""
		for c in range(0, len(img_as_ascii), new_width):
			#print img_as_ascii[c:c+new_width]
			output += img_as_ascii[c:c+new_width] + u'\n'
		return output



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	DodoLaSaumure
# Date:	30-Dec-2019

from __future__ import absolute_import
from __future__ import print_function
__author__ = "DodoLaSaumure"
__email__  = ""



import math
from bmath import Vector
from CNC import CW,CNC,Block
from ToolsPage import Plugin
from CNCRibbon    import Page
try:
	import Tkinter
	import tkMessageBox
except ImportError:
	import tkinter
	import tkinter.messagebox as tkMessageBox


#==============================================================================
# Create a simple Translate
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Translates a block to a new position")

	def __init__(self, master):
		Plugin.__init__(self, master, "SimpleTranslate")
		self.icon  = "SimpleTranslate"
		self.group = "Generator"
		self.variables = [
			("xinc",        "float",    10.0, _("x increment")),
			("yinc",        "float",    10.0, _("y increment")),
			("nbrepeat",        "int",    2, _("nb repeat including original")),
			("keep",        "bool",    True, _("Keep original Yes/No")),
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		n = self["name"]
		dx = self["xinc"]
		dy = self["yinc"]
		nbrepeat = self["nbrepeat"]
		if nbrepeat ==1 :
			nbrepeat =2
		keep = self["keep"]
		blocks = app.editor.getSelectedBlocks()
		if not blocks:
			app.editor.selectAll()
			blocks = app.editor.getSelectedBlocks()
		if not blocks:
			tkMessageBox.showerror(_("Tile error"),
				_("No g-code blocks selected"))
			return
		pos = blocks[-1]	# insert position
		y = dy
		x = dx
		pos += 1
		for index in range(int(nbrepeat-1)):
			# clone selected blocks
			undoinfo = []
			newblocks = []
			for bid in blocks:
				undoinfo.append(app.gcode.cloneBlockUndo(bid, pos))
				newblocks.append((pos,None))
				pos += 1
			app.addUndo(undoinfo)
			app.gcode.moveLines(newblocks, x, y)
			x += dx
			y += dy
		if not keep :
			app.editor.deleteBlock()
		app.refresh()
		app.setStatus(_("Moved selected blocks"))

if __name__=="__main__":
	pass
